@define-color livi_control_bg_color alpha(@window_bg_color, 0.9);

headerbar {
  background: none;
}

headerbar button {
  padding: 6px;
  background-color: @livi_control_bg_color;
}

windowcontrols button.close {
  background: @livi_control_bg_color;
  padding: 0;
  border-radius: 9999px;
  -gtk-icon-size: 16px;
  min-width: 48px;
}

windowcontrols button.close image {
  background: none;
}

.livi-window {
  background-color: black;
}

.livi-controls {
  background-color: @livi_control_bg_color;
  font-feature-settings: "tnum";
}

.livi-controls.wide {
  margin: 16px;
  border-radius: 6px;
}

.livi-error {
  background-color: rgba(80, 80, 80, 0.8);
  color: @warning_color;
  margin: 4px 16px 4px 16px;
  border-radius: 4px;
}

.livi-seek {
  -gtk-icon-shadow: 3px 3px 5px alpha(black, 0.75);
  text-shadow: 3px 3px 5px alpha(black, 0.75);
}
