/*
 * Copyright 2026 Phosh.mobi e.V.
 *
 * SPDX-License-Identifier: GPL-3.0-or-later
 */

#pragma once

#include <gtk/gtk.h>

G_BEGIN_DECLS

#define LIVI_TYPE_MARGIN_LAYOUT (livi_margin_layout_get_type())

G_DECLARE_FINAL_TYPE (LiviMarginLayout, livi_margin_layout, LIVI, MARGIN_LAYOUT, GtkLayoutManager)

GtkLayoutManager *livi_margin_layout_new (void);

G_END_DECLS
