/*
 * Copyright (C) 2026 Phosh.mobi e.V.
 *
 * SPDX-License-Identifier: GPL-3.0-or-later
 */

#pragma once

#include <glib-object.h>

G_BEGIN_DECLS

#define LIVI_TYPE_CUTOUTS ( livi_cutouts_get_type ())

G_DECLARE_FINAL_TYPE (LiviCoutouts, livi_cutouts, LIVI, LIVI_CUTOUTS, GObject)

LiviCutouts livi_cutouts_new (void);

G_END_DECLS
